# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[updateMap-all-options] 1'] = {
    'data': {
        'updateMap': {
            'map': {
                'mapId': '1001',
                'name': 'new-name'
            }
        }
    }
}

snapshots['test_schema_success[updateMap-all-options] 2'] = {
    'data': {
        'map': {
            'beams': {
                'edges': [
                ]
            },
            'datePosted': '2020-02-18',
            'mapFilePaths': {
                'edges': [
                    {
                        'node': {
                            'path': 'nersc:/go/to/my/maps'
                        }
                    }
                ]
            },
            'mapper': 'pwg-xyz',
            'name': 'new-name',
            'note': '- Note 123'
        }
    }
}

snapshots['test_schema_success[updateMap-selective-options] 1'] = {
    'data': {
        'updateMap': {
            'map': {
                'mapId': '1001',
                'name': 'new-name'
            }
        }
    }
}

snapshots['test_schema_success[updateMap-selective-options] 2'] = {
    'data': {
        'map': {
            'beams': {
                'edges': [
                ]
            },
            'datePosted': '2019-02-13',
            'mapFilePaths': {
                'edges': [
                    {
                        'node': {
                            'path': 'nersc:/go/to/my/maps'
                        }
                    }
                ]
            },
            'mapper': 'pwg-xyz',
            'name': 'new-name',
            'note': '''- This is a dummy test with a lat map
- This should not depend on any beam'''
        }
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots['test_export_db_to_csv_files 1'] = {
    'beam_file_paths': [
        {
            'beam_file_path_id': 1,
            'beam_id': 1070,
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20190304'
        },
        {
            'beam_file_path_id': 2,
            'beam_id': 1120,
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20190607'
        },
        {
            'beam_file_path_id': 3,
            'beam_id': 1130,
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20200123'
        },
        {
            'beam_file_path_id': 4,
            'beam_id': 1150,
            'note': '',
            'path': 'BEAM_DEPOT/Beams/20200207'
        }
    ],
    'beams': [
        {
            'beam_id': 1010,
            'input_beam_id': '',
            'input_map_id': '',
            'name': '20180101'
        },
        {
            'beam_id': 1070,
            'input_beam_id': '',
            'input_map_id': '',
            'name': '20190304'
        },
        {
            'beam_id': 1120,
            'input_beam_id': '',
            'input_map_id': '',
            'name': '20190607'
        },
        {
            'beam_id': 1130,
            'input_beam_id': '',
            'input_map_id': 1012,
            'name': '20200123'
        },
        {
            'beam_id': 1150,
            'input_beam_id': 1130,
            'input_map_id': 1013,
            'name': '20200207'
        }
    ],
    'map_file_paths': [
        {
            'map_file_path_id': 1,
            'map_id': 1001,
            'note': '',
            'path': 'nersc:/go/to/my/maps'
        },
        {
            'map_file_path_id': 2,
            'map_id': 1012,
            'note': 'lat only',
            'path': 'nersc:/go/to/my/maps_v2'
        },
        {
            'map_file_path_id': 3,
            'map_id': 1012,
            'note': 'lat only',
            'path': 'abcde:/path/to/the/maps_v2'
        },
        {
            'map_file_path_id': 4,
            'map_id': 1013,
            'note': 'lat only',
            'path': 'nersc:/go/to/my/maps_v3'
        }
    ],
    'maps': [
        {
            'date_posted': GenericRepr('datetime.date(2019, 2, 13)'),
            'map_id': 1001,
            'mapper': 'pwg-pmn',
            'name': 'lat20190213',
            'note': '''- This is a dummy test with a lat map
- This should not depend on any beam'''
        },
        {
            'date_posted': GenericRepr('datetime.date(2020, 1, 20)'),
            'map_id': 1012,
            'mapper': 'pwg-pmn',
            'name': 'lat20200120',
            'note': '''- This is a dummy test with a lat map
- A beam depends on this map'''
        },
        {
            'date_posted': GenericRepr('datetime.date(2020, 2, 1)'),
            'map_id': 1013,
            'mapper': 'pwg-pmn',
            'name': 'lat20200201',
            'note': '''- This is a dummy test with a lat map
- A beam depends on this map'''
        }
    ],
    'simulation_file_paths': [
        {
            'note': '',
            'path': 'nersc:/go/to/my/simulations',
            'simulation_file_path_id': 1,
            'simulation_id': 1001
        },
        {
            'note': '',
            'path': 'abcde:/path/to/the/simulations',
            'simulation_file_path_id': 2,
            'simulation_id': 1001
        }
    ],
    'simulations': [
        {
            'date_posted': GenericRepr('datetime.date(2019, 3, 15)'),
            'mapper': 'abc-def',
            'name': 'xyz-s1234-20200101',
            'note': '''- note 1
- note 2''',
            'simulation_id': 1001
        }
    ]
}



"""Commands (click)

"""
from flask.cli import with_appcontext
import click

import json

from .ops import define_tables
from .ops import export_db_to_dict_of_dict_list
from .ops import import_tables_from_csv_files
from .ops import export_db_to_csv_files
from .backup import backup_db

##__________________________________________________________________||
@click.command("init-db")
@with_appcontext
def init_db_command():
    """Clear existing data and create new tables."""
    define_tables()
    click.echo("Initialized the database.")

##__________________________________________________________________||
@click.command("dump-db")
@with_appcontext
def dump_db_command():
    """Dump the DB contents

    """
    db_content = export_db_to_dict_of_dict_list()

    click.echo(json.dumps(db_content, indent=2, default=str))
    # https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable

##__________________________________________________________________||
@click.command("import-csv")
@click.argument("csvdir", type=click.Path(exists=True))
@with_appcontext
def import_csv_command(csvdir):
    """Import tables from CSV files in CSVDIR into the DB.

    """
    import_tables_from_csv_files(csvdir)

##__________________________________________________________________||
@click.command("export-csv")
@click.argument("csvdir", type=click.Path())
@with_appcontext
def export_csv_command(csvdir):
    """Export the tables into the DB to CSV files in CSVDIR.

    """
    export_db_to_csv_files(csvdir)

##__________________________________________________________________||
@click.command("backup-db")
@with_appcontext
def backup_db_command():
    """Back up the DB as CSV to GitHub

    """
    backup_db()


##__________________________________________________________________||



from pathlib import Path
import git

import pytest
from graphene.test import Client


from acondbs.db.ops import export_db_to_csv_files
from acondbs.schema.schema import schema

from acondbs.db.backup import backup_db_as_csv_to_github

##__________________________________________________________________||
@pytest.fixture
def local_repo(app, tmpdir_factory):
    """
    """

    # create a local repo with CSV files
    local_folder = Path(tmpdir_factory.mktemp('backup'))
    with app.app_context():
        export_db_to_csv_files(local_folder)
    local_repo = git.Repo.init(local_folder)
    local_repo.git.add(A=True)
    local_repo.index.commit('initial commit')

    yield local_repo

@pytest.fixture
def remote_repo(local_repo, tmpdir_factory):
    """remote repo

    """

    # create a remote repo (bare repo)
    folder = Path(tmpdir_factory.mktemp('backup'))
    remote_repo = git.Repo.init(folder, bare=True)
    remote_url = remote_repo.git_dir

    # push the current branch of the local repo to the remote repo
    remote = local_repo.create_remote('origin', url=remote_url)
    branch_name = local_repo.active_branch.name
    remote.push(refspec='{}:{}'.format(branch_name, branch_name))
    local_repo.heads.master.set_tracking_branch(remote.refs.master)

    yield remote_repo

##__________________________________________________________________||
def test_backup_db_as_csv_to_github(app, local_repo, remote_repo):

    repo_path = local_repo.working_tree_dir
    head_sha_old = local_repo.head.commit.hexsha
    assert head_sha_old == remote_repo.head.commit.hexsha

    # change the DB content
    mutation = '''
          mutation m {
            createMap(input: {
              name: "map1",
              datePosted: "2020-02-20",
              mapper: "pwg-pmn",
              note: "- Item 1"
            }) { map { name } }
          }
        '''
    client = Client(schema)
    with app.app_context():
        client.execute(mutation)

    # take backup
    with app.app_context():
        backup_db_as_csv_to_github(repo_path)

    # assert
    assert not local_repo.is_dirty(untracked_files=True)
    head_sha_new = local_repo.head.commit.hexsha
    assert not head_sha_old == head_sha_new
    assert head_sha_new == remote_repo.head.commit.hexsha

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[deleteMap] 1'] = {
    'data': {
        'deleteMap': {
            'ok': True
        }
    }
}

snapshots['test_schema_success[deleteMap] 2'] = {
    'data': {
        'map': None
    }
}

snapshots['test_schema_error[deleteMap-error] 1'] = {
    'data': {
        'deleteMap': None
    },
    'errors': [
        {
            'locations': [
                {
                    'column': 11,
                    'line': 3
                }
            ],
            'message': "Class 'builtins.NoneType' is not mapped",
            'path': [
                'deleteMap'
            ]
        }
    ]
}

snapshots['test_schema_error[deleteMap-error] 2'] = {
    'data': {
        'allMaps': {
            'edges': [
                {
                    'node': {
                        'mapId': '1001',
                        'name': 'lat20190213'
                    }
                },
                {
                    'node': {
                        'mapId': '1012',
                        'name': 'lat20200120'
                    }
                },
                {
                    'node': {
                        'mapId': '1013',
                        'name': 'lat20200201'
                    }
                }
            ]
        }
    }
}



from pathlib import Path

from flask import Flask
from flask_cors import CORS

from . import _warnings

##__________________________________________________________________||
DEFAULT_CONFIG_DICT = dict(
    SECRET_KEY='dev',
    SQLALCHEMY_DATABASE_URI="sqlite:///:memory:",
    SQLALCHEMY_TRACK_MODIFICATIONS=False
)

##__________________________________________________________________||
def create_app(config_path=None, **kwargs):

    app = Flask(__name__, instance_relative_config=False)
    app.config.from_mapping(**DEFAULT_CONFIG_DICT)

    if config_path is not None:
        config_path = Path(config_path)
        if not config_path.is_absolute():
            # If `config_path` is relative, it is considered relative
            # to the current working directory.
            config_path = Path.cwd().joinpath(config_path)

            # Note: `app.config.from_pyfile()` treats a relative path
            # either a) relative to the top directory of the app,
            # i.e., the directory in which this file is, if
            # `instance_relative_config` is `False`, or b) relative to
            # the "instance folder", a specific directory described in
            # https://flask.palletsprojects.com/en/1.1.x/config/#instance-folders),
            # if `instance_relative_config` is `True`.

        app.config.from_pyfile(config_path, silent=False)

    if kwargs:
        app.config.from_mapping(**kwargs)

    from .db import db
    db.init_app(app)

    from . import bpgraphql
    bpgraphql.init_app(app)

    CORS(app, resources={r'/*': {'origins': '*'}})

    return app

##__________________________________________________________________||

from ._version import get_versions

__version__ = get_versions()['version']
"""str: version

The version string, e.g., "0.1.2", "0.1.2+83.ga093a20.dirty".
generated from git tags by versioneer.

Versioneer: https://github.com/warner/python-versioneer

"""

del get_versions

##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import Map

# __________________________________________________________________||
def test_simple(app):
    '''A simple test of updating an object
    '''

    with app.app_context():
        map1 = Map.query.filter_by(map_id=1012).first()
        assert 'lat20200120' == map1.name
        map1.name = 'new-map-name'
        sa.session.commit()

    with app.app_context():
        map1 = Map.query.filter_by(map_id=1012).first()
        assert 'new-map-name' == map1.name

# __________________________________________________________________||



from acondbs.models import Map, Beam

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.

# __________________________________________________________________||
def test_backref(app):

    # the operations here don't need be within the app context.

    # 1. create a map
    map1 = Map(name="map1")

    # no beams are associated with the map
    assert [] == map1.beams


    # 2. create a beam with the map
    beam1 = Beam(name="beam1", map=map1)

    # the map is associated with the beam
    assert map1 is beam1.map

    # the beam is referenced back by the map
    assert [beam1] == map1.beams

    # the beam has no child beam
    assert 0 == len(beam1.child_beams)


    # 3. create another beam with the map and the beam
    beam2 = Beam(name="beam2", map=map1, parent_beam=beam1)

    # the map is also associated with the 2nd beam
    assert map1 is beam2.map

    # the two beams are referenced back by the map
    assert [beam1, beam2] == map1.beams

    # the 1st beam is referenced by the 2nd beam
    assert beam1 is beam2.parent_beam

    # the 2nd beam is referenced back by the 1st beam
    assert [beam2] == beam1.child_beams

# __________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[createMap-all-options] 1'] = {
    'data': {
        'createMap': {
            'map': {
                'name': 'map1'
            }
        }
    }
}

snapshots['test_schema_success[createMap-all-options] 2'] = {
    'data': {
        'map': {
            'beams': {
                'edges': [
                ]
            },
            'datePosted': '2020-02-20',
            'mapFilePaths': {
                'edges': [
                ]
            },
            'mapper': 'pwg-pmn',
            'name': 'map1',
            'note': '- Item 1'
        }
    }
}

snapshots['test_schema_success[createMap-selective-options] 1'] = {
    'data': {
        'createMap': {
            'map': {
                'name': 'map1'
            }
        }
    }
}

snapshots['test_schema_success[createMap-selective-options] 2'] = {
    'data': {
        'map': {
            'beams': {
                'edges': [
                ]
            },
            'datePosted': None,
            'mapFilePaths': {
                'edges': [
                ]
            },
            'mapper': 'pwg-pmn',
            'name': 'map1',
            'note': None
        }
    }
}

snapshots['test_schema_error[createMap-error-no-name] 1'] = {
    'errors': [
        {
            'locations': [
                {
                    'column': 30,
                    'line': 3
                }
            ],
            'message': '''Argument "input" has invalid value {mapper: "pwg-pmn"}.
In field "name": Expected "String!", found null.'''
        }
    ]
}

snapshots['test_schema_error[createMap-error-no-name] 2'] = {
    'data': {
        'map': None
    }
}



import pytest
import sqlalchemy

from acondbs.db.conn import get_db_connection

##__________________________________________________________________||
def test_get_close_db_connection(app):
    with app.app_context():
        conn = get_db_connection()
        assert conn is get_db_connection()

    with pytest.raises(sqlalchemy.exc.StatementError) as e:
        conn.execute("SELECT 1")

    assert "closed" in str(e.value)

##__________________________________________________________________||



import pytest
from graphene.test import Client

from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    pytest.param(
        '''
          mutation m {
            updateMapFilePath(mapFilePathId: 1, input: {
              path: "nersc:/go/to/my/new_map_v2",
              note: "- Note 1 updated",
              mapId: 1012
            }) { mapFilePath { path } }
          }
        ''',
        '''
          {
            map(mapId: 1012) {
              name datePosted mapper note
              beams { edges { node { name } } }
              mapFilePaths { edges { node { path note map { mapId } } } }
            }
          }
        ''',
        id='updateMapFilePath'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(mutation))
    with app.app_context():
        snapshot.assert_match(client.execute(query))
    assert 1 == mock_request_backup_db.call_count

##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import Map, Beam

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.

# __________________________________________________________________||
def test_simple(app):
    '''A simple test of adding an object
    '''

    with app.app_context():

        # save the initial number of the maps to compare later
        nmaps = len(Map.query.all())

    # this instantiation doesn't need be within a app context
    map1 = Map(name="map1")

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

    with app.app_context():

        # test the number of the maps is increased by one
        assert (nmaps + 1) == len(Map.query.all())

        # the new map can be retrieved in a different app context
        map1_ = Map.query.filter_by(name='map1').first()
        assert isinstance(map1_, Map)

# __________________________________________________________________||
def test_python_object(app):
    '''A simple test about Python object
    '''

    map1 = Map(name="map1")

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

        map1_ = Map.query.filter_by(name='map1').first()

        # the query returns the same Python object
        assert map1 is map1_

    with app.app_context():
        map1_ = Map.query.filter_by(name='map1').first()

        # In a different app context, no longer the same Python object
        assert map1 is not map1_

# __________________________________________________________________||
def test_primary_key(app):
    '''A simple test about the primary key
    '''

    map1 = Map(name="map1")

    # The primary key (map_id) is None at this point
    assert map1.map_id is None

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

        # After the commit, map_id is automatically assigned
        map_id = map1.map_id
        assert map_id is not None

    with app.app_context():

        # The object can be retrived by the map_id in another context
        map1 = Map.query.filter_by(map_id=map_id).first()
        assert 'map1' == map1.name

# __________________________________________________________________||
def test_relation(app):
    '''A simple test of adding an object with relation
    '''

    map1 = Map(name="map1")
    beam1 = Beam(name="beam1", map=map1)

    # The relation has been already established
    assert map1 is beam1.map
    assert [beam1] == map1.beams

    # The primary and foreign keys are still None
    assert map1.map_id is None
    assert beam1.beam_id is None
    assert beam1.input_map_id is None

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

        # The primary keys are assigned
        assert map1.map_id is not None
        assert beam1.beam_id is not None

        # The foreign key is correctly set
        assert map1.map_id == beam1.input_map_id

    with app.app_context():
        map1 = Map.query.filter_by(name='map1').first()
        beam1 = Beam.query.filter_by(name='beam1').first()

        # The relation is preserved in a different app context
        assert map1 is beam1.map
        assert beam1 is map1.beams[0]
        assert map1.map_id == beam1.input_map_id

# __________________________________________________________________||



import pytest
from graphene.test import Client

from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    pytest.param(
        '''
        { allSimulations {
             edges { node { name } }
           } }
         ''',
        id='allSimulations'
    ),
    pytest.param(
        '''
        { simulation(simulationId: 1001) { name } }
         ''',
        id='simulationBySimulationID'
    ),
    pytest.param(
        '''
        { simulation(simulationId: 2001) { name } }
         ''',
        id='simulationBySimulationID-nonexistent'
    ),
]

@pytest.mark.parametrize('query', params)
def test_schema(app, snapshot, query):
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(query))

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[beamByBeamID] 1'] = {
    'data': {
        'beam': {
            'name': '20180101'
        }
    }
}

snapshots['test_schema[beamByBeamID-nonexistent] 1'] = {
    'data': {
        'beam': None
    }
}

snapshots['test_schema[beamByName] 1'] = {
    'data': {
        'beam': {
            'beamId': '1010'
        }
    }
}



import csv

import pytest

##__________________________________________________________________||
multiline_entry = """\
"abc
xyz",def,ghi\
"""

params = [
    pytest.param(
        ('abc,def,ghi', ),
        [['abc','def','ghi']],
        id='simple'
    ),

    pytest.param(
        ('"abc",def,ghi', ),
        [['abc','def','ghi']],
        id='double-quotes-removed'
    ),

    pytest.param(
        ("'abc',def,ghi", ),
        [["'abc'",'def','ghi']],
        id='single-quotes-not-removed'
    ),

    pytest.param(
        ('"""abc",def,ghi', ),
        [['"abc','def','ghi']],
        id='to-include-double-quote'),

    pytest.param(
        ('"ab,c",def,ghi', ),
        [['ab,c','def','ghi']],
        id='to-include-comma'
    ),

    pytest.param(
        (multiline_entry, ),
        [['abc\nxyz','def','ghi']],
        id='to-include-linebreak'
    ),
]

@pytest.mark.parametrize('input, expected', params)
def test_csv_specification(input, expected):
    """test CSV specification

    This test is used to check the CSV specification as implemented in
    csv.reader(), in particular, to understand how quotes are treated
    and how to include double quotes, commas, and line breaks in data

    """
    rows = list(csv.reader(input))
    assert expected == rows

##__________________________________________________________________||



import pytest
from graphene.test import Client

from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    pytest.param(
        '''
          mutation m {
            createMap(input: {
              name: "map1",
              datePosted: "2020-02-20",
              mapper: "pwg-pmn",
              note: "- Item 1"
            }) { map { name } }
          }
        ''',
        '''
          {
            map(name: "map1") {
              name datePosted mapper note
              beams { edges { node { name } } }
              mapFilePaths { edges { node { path } } }
            }
          }
        ''',
        id='createMap-all-options'
    ),
    pytest.param(
        '''
          mutation m {
            createMap(input: {
              name: "map1",
              mapper: "pwg-pmn"
            }) { map { name } }
          }
        ''',
        '''
          {
            map(name: "map1") {
              name datePosted mapper note
              beams { edges { node { name } } }
              mapFilePaths { edges { node { path } } }
            }
          }
        ''',
        id='createMap-selective-options'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(mutation))
    with app.app_context():
        snapshot.assert_match(client.execute(query))
    assert 1 == mock_request_backup_db.call_count

##__________________________________________________________________||
params = [
    pytest.param(
        '''
          mutation m {
            createMap(input: {
              mapper: "pwg-pmn"
            }) { map { name } }
          }
        ''',
        '''
          {
            map(name: "map1") {
              name datePosted mapper note
              beams { edges { node { name } } }
              mapFilePaths { edges { node { path } } }
            }
          }
        ''',
        id='createMap-error-no-name'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(mutation))
    with app.app_context():
        snapshot.assert_match(client.execute(query))
    assert 0 == mock_request_backup_db.call_count

##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import Map, Beam

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.

# __________________________________________________________________||
def test_simple(app):
    '''A simple test of deleting an object
    '''

    with app.app_context():

        # save the initial number of the maps to compare later
        nmaps = len(Map.query.all())

    with app.app_context():
        map1 = Map.query.filter_by(map_id=1012).first()
        sa.session.delete(map1)
        sa.session.commit()

    with app.app_context():

        # test the number of the maps is decreased by one
        assert (nmaps - 1) == len(Map.query.all())

        # the map is no longer found
        map1 = Map.query.filter_by(map_id=1012).first()
        assert map1 is None

# __________________________________________________________________||



import graphene

from .map_ import CreateMap, UpdateMap, DeleteMap
from .map_file_path import CreateMapFilePath, UpdateMapFilePath, DeleteMapFilePath

from .query import Query

##__________________________________________________________________||
class Mutation(graphene.ObjectType):
    create_map = CreateMap.Field()
    update_map = UpdateMap.Field()
    delete_map = DeleteMap.Field()

    create_map_file_path = CreateMapFilePath.Field()
    update_map_file_path = UpdateMapFilePath.Field()
    delete_map_file_path = DeleteMapFilePath.Field()

##__________________________________________________________________||



"""Backup DB
"""
from pathlib import Path
import atexit
import threading
import subprocess
import warnings

from flask import current_app

from acondbs.misc.cap import cap_exec_rate
from acondbs.db.ops import export_db_to_csv_files
from acondbs.misc import gitb
from acondbs.misc import lock

##__________________________________________________________________||
def request_backup_db():
    global _lock
    global _capped_backup_func
    with _lock:
        if not _capped_backup_func:
            pause = current_app.config['ACONDBS_DB_BACKUP_CSV_GIT_PAUSE']
            _capped_backup_func = cap_exec_rate(
                func=run_flask_backup_db,
                pause_time=pause, daemon=True)
                # Need to set daemon=True. Otherwise
                # threading waits for the thread to join
                # before the function registered in atexist
                # is executed.
    _capped_backup_func()

_lock = threading.Lock()
_capped_backup_func = None

##__________________________________________________________________||
def end_backup_thread():
    global _lock
    global _capped_backup_func
    with _lock:
        if _capped_backup_func:
            _capped_backup_func.end()
        _capped_backup_func = None

import multiprocessing.queues # This import prevents the error described in
                               # https://github.com/alphatwirl/atpbar/issues/4#issuecomment-473426630

atexit.register(end_backup_thread)

##__________________________________________________________________||
def run_flask_backup_db():
    proc = subprocess.run(['flask', 'backup-db'])

##__________________________________________________________________||
def backup_db():
    repo_path = current_app.config['ACONDBS_DB_BACKUP_CSV_GIT_FOLDER']
    lock_path = current_app.config['ACONDBS_DB_BACKUP_CSV_GIT_LOCK']
    timeout = current_app.config['ACONDBS_DB_BACKUP_CSV_GIT_LOCK_TIMEOUT']
    try:
        with lock.lock(lock_path, timeout=timeout):
            backup_db_as_csv_to_github(repo_path)
    except lock.TimeOutAcquiringLock:
        warnings.warn('Time out! unable to acquire the lock in {} seconds: {}'.format(timeout, lock_path))

##__________________________________________________________________||
def backup_db_as_csv_to_github(repo_path):
    repo_path = Path(repo_path)
    for csv_file in repo_path.glob('*.csv'):
        csv_file.unlink()
    export_db_to_csv_files(repo_path)
    gitb.commit(repo_path)
    gitb.push(repo_path)

##__________________________________________________________________||



import pytest
from graphene.test import Client

from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    pytest.param(
        '''
        { beam(beamId: 1010) { name } }
         ''',
        id='beamByBeamID'
    ),
    pytest.param(
        '''
        { beam(beamId: 2001) { name } }
         ''',
        id='beamByBeamID-nonexistent'
    ),
    pytest.param(
        '''
        { beam(name: "20180101") { beamId } }
         ''',
        id='beamByName'
    ),
]

@pytest.mark.parametrize('query', params)
def test_schema(app, snapshot, query):
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(query))

##__________________________________________________________________||



from pathlib import Path

import pytest
import unittest.mock as mock

from ..constants import SAMPLE_DIR

##__________________________________________________________________||
@pytest.fixture()
def mock_define_tables(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr("acondbs.db.cmds.define_tables", ret)
    return ret

def test_init_db_command(runner, mock_define_tables):
    """test command init-db
    """
    result = runner.invoke(args=["init-db"])
    assert [mock.call()] == mock_define_tables.call_args_list
    assert "Initialized" in result.output

##__________________________________________________________________||
def test_dump_db_command(runner):
    """test command dump-db
    """
    result = runner.invoke(args=["dump-db"])
    assert 0 == result.exit_code
    assert 1800 < len(result.output)

##__________________________________________________________________||
@pytest.fixture()
def mock_import_tables_from_csv_files(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr("acondbs.db.cmds.import_tables_from_csv_files", ret)
    return ret

def test_import_csv_command(runner, mock_import_tables_from_csv_files):
    """test command import-csv
    """

    csvdir = str(Path(SAMPLE_DIR, 'csv'))
    # needs to give an existing path to `import-csv` as `click.Path()`
    # checks the check the existence.

    result = runner.invoke(args=["import-csv", csvdir])
    assert 0 == result.exit_code
    assert [mock.call(csvdir)] == mock_import_tables_from_csv_files.call_args_list

##__________________________________________________________________||
@pytest.fixture()
def mock_export_db_to_csv_files(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr("acondbs.db.cmds.export_db_to_csv_files", ret)
    return ret

def test_export_csv_command(runner, tmpdir_factory, mock_export_db_to_csv_files):
    """test command export-csv
    """

    tmpdir = str(tmpdir_factory.mktemp('csv_out'))
    csvdir = str(Path(tmpdir, 'csv'))

    result = runner.invoke(args=["export-csv", csvdir])
    assert 0 == result.exit_code
    assert [mock.call(csvdir)] == mock_export_db_to_csv_files.call_args_list

##__________________________________________________________________||



import datetime
import sqlalchemy
import pytest

from acondbs.db.sa import sa
from acondbs.models import Map

# __________________________________________________________________||
def test_type(app):
    '''confirm the type of the date field
    '''

    with app.app_context():
        map = Map.query.filter_by(name='lat20200120').first()

        # The type of the field "date_posted" of Map is "datetime.date"
        assert isinstance(map.date_posted, datetime.date)

# __________________________________________________________________||
def test_add(app):
    '''A simple test of adding an object with a date field
    '''

    # date_posted needs to be initialized with a datetime.date
    date_posted = datetime.date(2019, 2, 23)
    map1 = Map(name="map1", date_posted=date_posted)

    with app.app_context():
        sa.session.add(map1)
        sa.session.commit()

    with app.app_context():
        map1 = Map.query.filter_by(name='map1').first()
        assert datetime.date(2019, 2, 23) == map1.date_posted

# __________________________________________________________________||
def test_add_raise(app):
    '''A simple test of adding an object with a wrong type
    '''

    # It is not impossible to instnaiate a date field with a wrong
    # type, e.g, str
    map1 = Map(name="map1", date_posted="2019-02-13")

    with app.app_context():

        # It is also possible to add
        sa.session.add(map1)

        # However, it is not possible to commit
        with pytest.raises(sqlalchemy.exc.StatementError):
            sa.session.commit()

# __________________________________________________________________||



import pytest
from graphene.test import Client

import acondbs
from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    pytest.param(
        '{ version }',
        {'version': acondbs.__version__},
        id='version'
    ),
]

@pytest.mark.parametrize('query, expected', params)
def test_schema(app, query, expected):
    with app.app_context():
        client = Client(schema)
        result = client.execute(query)
        assert {'data': expected} == result

##__________________________________________________________________||



from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Simulation as SimulationModel

##__________________________________________________________________||
class Simulation(SQLAlchemyObjectType):
    class Meta:
        model = SimulationModel
        interfaces = (relay.Node, )

##__________________________________________________________________||



import os
from pathlib import Path
import time
import contextlib

##__________________________________________________________________||
class lock:

    def __init__(self, path, timeout=None):
        self.path = Path(path)
        self.timeout = timeout
        self.locked = False

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def acquire(self):
        start_time = time.time()
        while True:
            if try_make_file(self.path):
                break
            if self.timeout is not None:
                if time.time() - start_time > self.timeout:
                    raise TimeOutAcquiringLock
            time.sleep(0.001)
        self.locked = True

    def release(self):
        if not self.locked:
            return

        try:
            self.path.unlink()
        except FileNotFoundError:
            pass
        self.locked = False

##__________________________________________________________________||
class TimeOutAcquiringLock(Exception):
    pass

##__________________________________________________________________||
def try_make_file(path):
    """try to create a file atomically

    http://stackoverflow.com/questions/33223564/atomically-creating-a-file-if-it-doesnt-exist-in-python
    """
    try:
        os.open(path,  os.O_CREAT | os.O_EXCL)
        return True
    except FileExistsError:
        return False

##__________________________________________________________________||



"""Git operations
"""
from pathlib import Path
import git
import warnings

##__________________________________________________________________||
def commit(path, message=None):
    """commit all changes in a git repository to git

    If the path is not a git repository, unless it is empty, it will
    be initialized as a new repository.

    Parameters
    ----------
    path : str
        The path to a git repository
    message : str, optional
        Commit message.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If `path` is not a directory

    """

    path = Path(path)
    if not path.is_dir():
        raise ValueError("path is not a directory: {}".format(path))

    if is_git_repo(path):
        repo = git.Repo(path)
    else:
        if not list(path.iterdir()):
            warnings.warn('the directory is empty. nothing to commit: {}'.format(path))
            return
        repo = git.Repo.init(path)

    if not repo.is_dirty(untracked_files=True):
        warnings.warn('the repo is clean. nothing to commit: {}'.format(path))
        return

    repo.git.add(A=True)

    if not message:
        message = 'commit all'
    repo.index.commit(message)

##__________________________________________________________________||
def pull(path):
    """pull from a tracking branch

    Equivalent to execute "git pull" in the path.

    Parameters
    ----------
    path : str
        The path to a git repository

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If `path` is not a directory
        If `path` is not a git repo
        If the repo has no tracking branch

    """

    path = Path(path)
    if not path.is_dir():
        raise ValueError("path is not a directory: {}".format(path))

    if not is_git_repo(path):
        raise ValueError("path is not a git repo: {}".format(path))

    repo = git.Repo(path)

    tracking_branch = repo.active_branch.tracking_branch()

    if not tracking_branch:
        raise ValueError("repo has no tracking branch: {}".format(path))

    remote = repo.remotes[tracking_branch.remote_name]
    remote.pull()

##__________________________________________________________________||
def push(path):
    """push to a upstream branch

    Equivalent to execute "git push" in the path.

    Parameters
    ----------
    path : str
        The path to a git repository

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If `path` is not a directory
        If `path` is not a git repo
        If the repo has no tracking branch

    """

    path = Path(path)
    if not path.is_dir():
        raise ValueError("path is not a directory: {}".format(path))

    if not is_git_repo(path):
        raise ValueError("path is not a git repo: {}".format(path))

    repo = git.Repo(path)

    tracking_branch = repo.active_branch.tracking_branch()

    if not tracking_branch:
        raise ValueError("repo has no tracking branch: {}".format(path))

    remote = repo.remotes[tracking_branch.remote_name]
    remote.push()

##__________________________________________________________________||
def is_git_repo(path):
    """test if a folder is a git repository

    copied from https://stackoverflow.com/a/39956572/7309855

    """
    try:
        _ = git.Repo(path).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False

##__________________________________________________________________||



from acondbs.db.sa import sa

# __________________________________________________________________||
def test_models(app):
    '''test the models declared
    '''
    expected = {
        'simulations', 'simulation_file_paths',
        'maps', 'map_file_paths',
        'beams', 'beam_file_paths'
    }
    model_names = sa.Model.metadata.tables.keys()
    assert expected == model_names

# __________________________________________________________________||



import git

from pathlib import Path
import warnings

import pytest

from acondbs.misc import gitb

##__________________________________________________________________||
def test_empty_folder(tmpdir_factory):
    """assert empty folder won't be initialized as a repo
    """
    folder = Path(tmpdir_factory.mktemp('git'))
    with warnings.catch_warnings(record=True) as w:
        gitb.commit(folder)
    assert len(w) == 1
    assert not gitb.is_git_repo(folder)

##__________________________________________________________________||
@pytest.fixture()
def nonexistent_path(tmpdir_factory):
    """path to a nonexistent file
    """
    folder = Path(tmpdir_factory.mktemp('git'))
    path = folder.joinpath('nonexistent')
    yield path

def test_nonexistent_path(nonexistent_path):
    """assert exception is raised for nonexistent path
    """
    with pytest.raises(ValueError):
        gitb.commit(nonexistent_path)

##__________________________________________________________________||
def test_non_empty_folder(folder):
    """assert a repo initialized and files committed
    """
    gitb.commit(folder)
    repo = git.Repo(folder)
    assert not repo.is_dirty(untracked_files=True)
    ncommits = len(list(repo.iter_commits()))
    assert 1 == ncommits

##__________________________________________________________________||
def test_clean_repo(repo):
    """assert no empty commit is made
    """
    folder = repo.working_tree_dir
    with warnings.catch_warnings(record=True) as w:
        gitb.commit(folder)
    assert len(w) == 1
    assert not repo.is_dirty(untracked_files=True)
    ncommits = len(list(repo.iter_commits()))
    assert 1 == ncommits

@pytest.fixture()
def repo_dirty(repo):
    """a dirty repo

    The repo cotains a staged file, a modified file, a deleted file,
    and an untracked file.

    """
    folder = Path(repo.working_tree_dir)
    file1 = folder.joinpath('f.txt')
    file2 = folder.joinpath('g.txt')
    file3 = folder.joinpath('h.txt')
    file1.write_text('xyz')
    repo.git.add(file1)
    file1.write_text('zzz')
    file2.unlink()
    file3.write_text('qwe')
    yield repo

def test_dirty_repo(repo_dirty):
    """assert all changes commit
    """
    repo = repo_dirty
    folder = repo.working_tree_dir
    gitb.commit(folder)
    assert not repo.is_dirty(untracked_files=True)
    ncommits = len(list(repo.iter_commits()))
    assert 2 == ncommits

def test_default_message(repo_dirty):
    repo = repo_dirty
    folder = repo.working_tree_dir
    gitb.commit(folder)
    assert 'commit all' == repo.head.commit.message

def test_custom_message(repo_dirty):
    repo = repo_dirty
    folder = repo.working_tree_dir
    gitb.commit(folder, 'custom message')
    assert 'custom message' == repo.head.commit.message

##__________________________________________________________________||



import time
from acondbs.misc.cap import cap_exec_rate
from acondbs.misc.cap import State

import pytest
import unittest.mock as mock

##__________________________________________________________________||
class Task:
    def __init__(self):
        self.counter = 0

    def __call__(self):
        self.counter += 1

def test_init():
    task = Task()
    cap = cap_exec_rate(func=task, pause_time=0.1)
    assert cap.func is task
    assert 0.1 == cap.pause_time
    assert not cap.daemon
    cap.end()
    assert 0 == task.counter

def test_daemon():
    task = Task()
    cap = cap_exec_rate(func=task, pause_time=0.1, daemon=True)
    assert cap.daemon
    # end() doesn't need to be called
    assert 0 == task.counter

def test_call():
    task = Task()
    cap = cap_exec_rate(func=task, pause_time=0.1)
    cap()
    time.sleep(0.05)
    assert 1 == task.counter
    cap.end()

def test_call_2():
    task = Task()
    cap = cap_exec_rate(func=task, pause_time=0.05)
    cap()
    time.sleep(0.2)
    cap()
    time.sleep(0.05)
    assert 2 == task.counter
    cap.end()

def test_call_3():
    task = Task()
    cap = cap_exec_rate(func=task, pause_time=0.1)
    cap() # executed immediately
    cap() # not executed
    cap() # not executed
    cap() # not executed
    cap() # executed after the pause
    time.sleep(0.05)
    assert 1 == task.counter
    time.sleep(0.1)
    assert 2 == task.counter
    cap.end()

def test_call_4():
    task = Task()
    cap = cap_exec_rate(func=task, pause_time=10)
    cap() # executed immediately
    cap() # not executed
    cap() # not executed
    cap() # not executed
    cap() # executed at the end
    time.sleep(0.05)
    assert 1 == task.counter
    cap.end()
    assert 2 == task.counter

##__________________________________________________________________||
def test_state_raise():
    config = mock.Mock()
    config.queue.get().return_value = 'unknown message'
    state = State(config)
    with pytest.raises(ValueError):
        state()

##__________________________________________________________________||




import pytest
import unittest.mock as mock

from acondbs.db.backup import request_backup_db, run_flask_backup_db

##__________________________________________________________________||
@pytest.fixture()
def mock_cap_exec_rate(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.db.backup.cap_exec_rate", y)
    yield y

##__________________________________________________________________||
def test_request_backup_db(app, mock_cap_exec_rate):
    with app.app_context():
        pause = app.config['ACONDBS_DB_BACKUP_CSV_GIT_PAUSE']

    nrequests = 3
    with app.app_context():
        for i in range(nrequests):
            request_backup_db()

    assert 1 == mock_cap_exec_rate.call_count
    assert [
        mock.call(
            func=run_flask_backup_db, pause_time=pause, daemon=True
        )] == mock_cap_exec_rate.call_args_list

    assert nrequests == mock_cap_exec_rate().call_count

##__________________________________________________________________||



"""declare ORM models

In this module, ORM (Object-relational mapping) models are declared.
One model is mapped to one table in the DB. Models are declared as
Python classes inheriting the Model class in Flask-SQLAlchemy.

"Declaring Models" in Flask-SQLAlchemy doc:
https://flask-sqlalchemy.palletsprojects.com/en/2.x/models/

"Declare a Mapping" in SQLAlchemy doc:
https://docs.sqlalchhttps://docs.sqlalchemy.org/en/13/orm/tutorial.html#declare-a-mapping

"""

from .db.sa import sa

##__________________________________________________________________||
class Simulation(sa.Model):
    __tablename__ = 'simulations'
    simulation_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    date_posted = sa.Column(sa.Date())
    mapper = sa.Column(sa.Text())
    note = sa.Column(sa.Text())

class SimulationFilePath(sa.Model):
    __tablename__ = 'simulation_file_paths'
    simulation_file_path_id = sa.Column(sa.Integer(), primary_key=True)
    simulation_id = sa.Column(sa.ForeignKey('simulations.simulation_id'))
    path = sa.Column(sa.Text())
    note = sa.Column(sa.Text())
    simulation = sa.relationship("Simulation", backref=sa.backref("simulation_file_paths"))

class Map(sa.Model):
    __tablename__ = 'maps'
    map_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    date_posted = sa.Column(sa.Date())
    mapper = sa.Column(sa.Text())
    note = sa.Column(sa.Text())

class MapFilePath(sa.Model):
    __tablename__ = 'map_file_paths'
    map_file_path_id = sa.Column(sa.Integer(), primary_key=True)
    map_id = sa.Column(sa.ForeignKey('maps.map_id'))
    path = sa.Column(sa.Text())
    note = sa.Column(sa.Text())
    map = sa.relationship("Map", backref=sa.backref("map_file_paths"))

class Beam(sa.Model):
    __tablename__ = 'beams'
    beam_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    input_map_id = sa.Column(sa.ForeignKey('maps.map_id'))
    input_beam_id = sa.Column(sa.ForeignKey('beams.beam_id'))
    map = sa.relationship("Map", backref=sa.backref("beams"))
    parent_beam = sa.relationship(lambda: Beam, remote_side=beam_id, backref=sa.backref("child_beams"))

class BeamFilePath(sa.Model):
    __tablename__ = 'beam_file_paths'
    beam_file_path_id = sa.Column(sa.Integer(), primary_key=True)
    beam_id = sa.Column(sa.ForeignKey('beams.beam_id'))
    path = sa.Column(sa.Text())
    note = sa.Column(sa.Text())
    beam = sa.relationship("Beam", backref=sa.backref("beam_file_paths"))

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[allSimulations] 1'] = {
    'data': {
        'allSimulations': {
            'edges': [
                {
                    'node': {
                        'name': 'xyz-s1234-20200101'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[simulationBySimulationID] 1'] = {
    'data': {
        'simulation': {
            'name': 'xyz-s1234-20200101'
        }
    }
}

snapshots['test_schema[simulationBySimulationID-nonexistent] 1'] = {
    'data': {
        'simulation': None
    }
}



import pytest

from flask_sqlalchemy import BaseQuery

from acondbs.models import Map, Beam

# These tests are written primarily for the developer to understand
# how models in flask_sqlalchemy work.

# __________________________________________________________________||
def test_context(app):

    # query cannot be accessed outside of the app context
    with pytest.raises(RuntimeError):
        Map.query

    with app.app_context():
        Map.query

def test_query_all(app):

    with app.app_context():

        # query is an instance of BaseQuery
        query = Map.query
        assert isinstance(query, BaseQuery)

        # query.all() returns a list of maps
        results = query.all()
        # e.g., [<Map 1001>, <Map 1012>, <Map 1013>]

        assert ['lat20190213', 'lat20200120', 'lat20200201'] == [
            e.name for e in results]

        assert isinstance(results[0], Map)

def test_query_filter(app):

    with app.app_context():

        # filter_by() returns an instance of BaseQuery
        query = Map.query.filter_by(name='lat20200120')
        assert isinstance(query, BaseQuery)

        # the results are a list with one element
        results = query.all()
        assert 1 == len(results)
        assert ['lat20200120'] == [e.name for e in results]

        # first() returns a map
        map = query.first()
        assert isinstance(map, Map)
        assert 'lat20200120' == map.name

        # the back-referenced beam can be accessed as an attribute of
        # the map
        assert ['20200123'] == [e.name for e in map.beams]
        beam = map.beams[0]
        assert isinstance(beam, Beam)

        # the beam, of course, references to the map
        assert map is beam.map

def test_query_filter_nonexistent(app):

    with app.app_context():

        query = Map.query.filter_by(name='no-such-map')
        assert isinstance(query, BaseQuery)

        # the results are an empty list
        results = query.all()
        assert [] == results

        # first() returns None
        map = query.first()
        assert map is None

# __________________________________________________________________||



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import BeamFilePath as BeamFilePathModel

from ..db.sa import sa

##__________________________________________________________________||
class BeamFilePath(SQLAlchemyObjectType):
    class Meta:
        model = BeamFilePathModel
        interfaces = (relay.Node, )

##__________________________________________________________________||



"""Configure the module warnings

https://docs.python.org/3/library/warnings.html

"""
import warnings
from pathlib import Path

##__________________________________________________________________||
_module_path = Path(__file__).resolve().parent.parent
# the path to the dir in which the module is installed,
# i.e., the one dir above the module path.

def format(message, category, filename, lineno, file=None, line=None):
    filename = Path(filename).resolve().relative_to(_module_path)
    ret = '{}:{}: {}\n'.format(filename, lineno, message)
    return ret

warnings.formatwarning =format

##__________________________________________________________________||



from pathlib import Path
import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables, import_tables_from_csv_files, export_db_to_dict_of_dict_list

from ...constants import SAMPLE_DIR

##__________________________________________________________________||
@pytest.fixture
def app():
    database_uri ="sqlite:///:memory:"
    app = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with app.app_context():
        define_tables()
    yield app


##__________________________________________________________________||
def test_import_tables_from_csv_files(app, snapshot):
    csvdir = Path(SAMPLE_DIR, 'csv')
    with app.app_context():
        import_tables_from_csv_files(csvdir)
    with app.app_context():
        snapshot.assert_match(export_db_to_dict_of_dict_list())

##__________________________________________________________________||



from sqlalchemy import MetaData

import pytest

from acondbs.db.ops import export_db_to_dict_of_dict_list
from acondbs.db.sa import sa

##__________________________________________________________________||
def test_export_db_to_dict_of_dict_list(app, snapshot):
    """test export_db_to_dict_of_dict_list()

    """

    with app.app_context():
        snapshot.assert_match(export_db_to_dict_of_dict_list())

##__________________________________________________________________||



"""DB connection

The earlier versions of this module were similar to `db.py` in
`falskr`, an example application described in the flask tutorial.

 The flask tutorial: https://flask.palletsprojects.com/en/1.1.x/tutorial/

 db.py in flaskr: https://github.com/pallets/flask/blob/1.1.1/examples/tutorial/flaskr/db.py

"""
from flask import g

from .sa import sa

##__________________________________________________________________||
def get_db_connection():
    """returns the DB connection

    This function returns the `connection` of SQLAlchemy:
    https://docs.sqlalchemy.org/en/13/core/connections.html

    This function is similar to `get_db()` in `falskr`.

    """
    if 'db_connection' not in g:
        g.db_connection = sa.engine.connect()
    return g.db_connection

##__________________________________________________________________||
def close_db_connection(e=None):
    """closes the DB connection

    This function is registered to be called when the application ends.

    This function is similar to `close_db()` in `falskr`.
    """
    conn = g.pop('db_connection', None)
    if conn is not None:
        conn.close()

##__________________________________________________________________||



"""DB operations

The functions in this module need to be called within the application
context of Flask unless stated otherwise.

"""
import datetime
import csv
from pathlib import Path

from sqlalchemy import MetaData
import sqlalchemy

from .sa import sa
from .conn import get_db_connection

##__________________________________________________________________||
def define_tables():
    """defines tables in the DB

    This function defines tables in the DB after dropping all existing
    tables.

    """

    engine = sa.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)

    if metadata.tables:
        tbl_names = metadata.tables.keys()
        # ['maps', 'beams']

        tbl_names = ', '.join(['"{}"'.format(t) for t in tbl_names])
        # '"beams", "maps"'

        msg = "Dropped all tables: {}".format(tbl_names)

        metadata.drop_all(bind=engine)
        print(msg)

    if not sa.Model.metadata.tables:
        msg = "No tables to be created are found!"
        print(msg)
        return

    tbl_names = sa.Model.metadata.tables.keys()
    # ['maps', 'beams']

    tbl_names = ', '.join(['"{}"'.format(t) for t in tbl_names])
    # '"beams", "maps"'

    msg = "Created tables: {}".format(tbl_names)

    sa.Model.metadata.create_all(engine)

    print(msg)

##__________________________________________________________________||
def get_all_table_names():
    """returns the names of all tables in the DB.

    Returns
    -------
    list of str
        the names of all tables in the DB

    """
    engine = sa.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)
    return [tbl.name for tbl in metadata.sorted_tables]

def export_db_to_dict_of_dict_list():
    """exports the DB to a dict of table names and lists of dicts for each row

    Returns
    -------
    dict
        dict with one entry for each table: the table name as the key
        and the value lists of dicts: one dict for each row: field
        names as the keys and contents as the values. e.g.,

            {"tblA": [
                {'fieldA': 1001, 'fieldB': 'abc'},
                {'fieldA': 1002, 'fieldB': 'xyz'}
            ]}

    """
    tbl_names = get_all_table_names()
    ret = {n: export_table_to_dict_list(n) for n in tbl_names}
    return ret

def export_table_to_dict_list(tbl_name):
    """exports the table to a list of dicts

    Parameters
    ----------
    tbl_name : str
        the table name

    Returns
    -------
    list of dict
        the list of dicts: one dict for each row: field names as the
        keys and contents as the values. e.g.:

            [
                {'fieldA': 1001, 'fieldB': 'abc'},
                {'fieldA': 1002, 'fieldB': 'xyz'}
            ]

    """
    result_proxy = get_resultproxy_of_select_all_rows(tbl_name)
    return [dict(r) for r in result_proxy]

def get_resultproxy_of_select_all_rows(tbl_name):
    """returns ResultProxy with all rows of the table

    Parameters
    ----------
    tbl_name : str
        the table name

    Returns
    -------
    ResultProxy
        ResultProxy of SQLAlchemy with the result of query selecting
        all rows of the table.
        https://docs.sqlalchemy.org/en/13/core/connections.html#sqlalchemy.engine.ResultProxy

    """
    engine = sa.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)
    tbl = metadata.tables[tbl_name]
    return engine.execute(tbl.select())

##__________________________________________________________________||
def import_tables_from_csv_files(csvdir):
    """imports tables from CSV files into the DB

    The tables need to be already defined in the DB.

    The folder `csvdir` should contain CSV files with the table
    contents to be imported: one CSV file for one table with the file
    name <<table name>>.csv

    Parameters
    ----------
    csvdir : str
        a path to a folder with CSV files

    """
    tbl_names = get_all_table_names()
    for tbl_name in tbl_names:
        csv_filename = '{}.csv'.format(tbl_name)
        csv_path = Path(csvdir, csv_filename)
        if csv_path.exists():
            import_table_from_csv_file(tbl_name, csv_path)
            message = 'imported to "{}" from {}'.format(tbl_name, csv_path)
        else:
            message = 'skipped "{}". file not found: {}'.format(tbl_name, csv_path)
        print(message)

def import_table_from_csv_file(tbl_name, path):
    """import a table from a CSV file

    The table needs to be already defined in the DB.

    Parameters
    ----------
    tbl_name : str
        the name of the table
    path : str
        the path to the CSV file

    """
    engine = sa.engine
    metadata = MetaData()
    metadata.reflect(bind=engine)
    tbl = metadata.tables[tbl_name]

    with open(path, 'r') as f:
        rows = list(csv.reader(f))
    fields = rows[0]
    rows = rows[1:]
    data = [{f: convert_data_type_for_insert(e, tbl.columns[f].type)
             for f, e in zip(fields, r)} for r in rows]
    ins = tbl.insert()
    connection = get_db_connection()
    connection.execute(ins, data)

def convert_data_type_for_insert(str_, type_):
    """converts data type for insert

    This function converts the data type from str to a type relevant
    to insert in SQLAlchemy

    Parameters
    ----------
    str_ : str
        the data in str
    type_ :
        the data type in SQLAlchemy, i.e., one of the types listed in
        https://docs.sqlalchemy.org/en/13/core/type_basics.html

    Returns
    -------
    the data in the type relevant for insert

    """
    if isinstance(type_, sqlalchemy.sql.sqltypes.DATE):
        if str_:
            return datetime.datetime.strptime(str_, "%Y-%m-%d").date()
        return None
    return str_

##__________________________________________________________________||
def export_db_to_csv_files(outdir):
    """export all tables in the DB to CSV files

    The CSV files will be stored in the folder `csvdir`: one CSV file
    for one table with the file name <<table name>>.csv

    Parameters
    ----------
    csvdir : str
        a path to a folder to store the CSV files


    The tables need to be already defined in the DB.

    """
    tbl_names = get_all_table_names()

    Path(outdir).mkdir(parents=True, exist_ok=True)

    for tbl_name in tbl_names:
        csv_filename = '{}.csv'.format(tbl_name)
        csv_path = Path(outdir, csv_filename)
        result_proxy = get_resultproxy_of_select_all_rows(tbl_name)
        with open(csv_path, 'w', newline='') as f:
            csv_writer = csv.writer(f, lineterminator='\n')
            csv_writer.writerow(result_proxy.keys())
            csv_writer.writerows(result_proxy)

##__________________________________________________________________||



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Map as MapModel

from ..db.sa import sa
from ..db.backup import request_backup_db

##__________________________________________________________________||
class Map(SQLAlchemyObjectType):
    class Meta:
        model = MapModel
        interfaces = (relay.Node, )

class MapConnection(relay.Connection):
    class Meta:
        node = Map

##__________________________________________________________________||
class MapAttribute:
    name = graphene.String()
    date_posted = graphene.Date()
    mapper = graphene.String()
    note = graphene.String()

class CreateMapInput(graphene.InputObjectType, MapAttribute):
    name = graphene.String(required=True)

class UpdateMapInput(graphene.InputObjectType, MapAttribute):
    pass

class CreateMap(graphene.Mutation):
    class Arguments:
        input = CreateMapInput(required=True)

    ok = graphene.Boolean()
    map = graphene.Field(lambda: Map)

    def mutate(root, info, input):
        map = MapModel(**input)
        sa.session.add(map)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateMap(map=map, ok=ok)

class UpdateMap(graphene.Mutation):
    class Arguments:
        map_id = graphene.Int()
        input = UpdateMapInput(required=True)

    ok = graphene.Boolean()
    map = graphene.Field(lambda: Map)

    def mutate(root, info, map_id, input):
        map = MapModel.query.filter_by(map_id=map_id).first()
        for k, v in input.items():
            setattr(map, k, v)
        sa.session.commit()
        ok = True
        request_backup_db()
        return UpdateMap(map=map, ok=ok)

class DeleteMap(graphene.Mutation):
    class Arguments:
        map_id = graphene.Int()

    ok = graphene.Boolean()

    def mutate(root, info, map_id):
        map = MapModel.query.filter_by(map_id=map_id).first()
        sa.session.delete(map)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteMap(ok=ok)

##__________________________________________________________________||



from pathlib import Path
import warnings

import pytest
import unittest.mock as mock

from acondbs.db.backup import backup_db

##__________________________________________________________________||
@pytest.fixture()
def mock_backup_db_as_csv_to_github(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.db.backup.backup_db_as_csv_to_github", y)
    yield y

@pytest.fixture()
def mock_lock_path(app, monkeypatch, tmpdir_factory):
    folder = Path(tmpdir_factory.mktemp('backup'))
    y = folder.joinpath('lock')
    monkeypatch.setitem(app.config, 'ACONDBS_DB_BACKUP_CSV_GIT_LOCK', str(y))
    monkeypatch.setitem(app.config, 'ACONDBS_DB_BACKUP_CSV_GIT_LOCK_TIMEOUT', 0.01)
    yield y

##__________________________________________________________________||
def test_backup_db(app, mock_lock_path, mock_backup_db_as_csv_to_github):
    with app.app_context():
        repo_path = app.config['ACONDBS_DB_BACKUP_CSV_GIT_FOLDER']
        backup_db()
    assert [mock.call(repo_path)] == mock_backup_db_as_csv_to_github.call_args_list

##__________________________________________________________________||
def test_backup_db_locked(app, mock_lock_path, mock_backup_db_as_csv_to_github):
    mock_lock_path.touch()
    with app.app_context():
        with warnings.catch_warnings(record=True) as w:
            backup_db()
    assert [] == mock_backup_db_as_csv_to_github.call_args_list
    assert len(w) == 1

##__________________________________________________________________||



from sqlalchemy import MetaData

import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables
from acondbs.db.sa import sa

##__________________________________________________________________||
@pytest.fixture
def app_with_empty_db():
    database_uri ="sqlite:///:memory:"
    app = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    yield app

##__________________________________________________________________||
def test_define_tables_start_with_empty_db(app_with_empty_db):
    """test define_tables()

    This function tests if tables will be defined starting from new db without
    any tables.

    """

    app = app_with_empty_db

    # confirm tables are not defined initially
    with app.app_context():
        metadata = MetaData()
        metadata.reflect(bind=sa.engine)
        assert not metadata.tables

    with app.app_context():
        define_tables()

    with app.app_context():
        metadata = MetaData()
        metadata.reflect(bind=sa.engine)
        tbl_names = {
            'simulations', 'simulation_file_paths',
            'maps', 'map_file_paths',
            'beams', 'beam_file_paths'
        }
        assert tbl_names == metadata.tables.keys()

##__________________________________________________________________||
def test_define_tables_start_with_nonempty_db(app):
    """test define_tables()

    This function tests if tables will be redefined starting from db
    with tables with entries.

    """

    # confirm tables are defined initially and not empty
    with app.app_context():
        metadata = MetaData()
        metadata.reflect(bind=sa.engine)
        assert metadata.tables
        total_nentries = sum([len([r for r in
                               sa.engine.execute(tbl.select())]) for tbl in
                          metadata.sorted_tables])
        assert total_nentries > 0

    with app.app_context():
        define_tables()

    # confirm tables are defined and all empty
    with app.app_context():
        metadata = MetaData()
        metadata.reflect(bind=sa.engine)
        tbl_names = {
            'simulations', 'simulation_file_paths',
            'maps', 'map_file_paths',
            'beams', 'beam_file_paths'
        }
        assert tbl_names == metadata.tables.keys()
        total_nentries = sum([len([r for r in
                               sa.engine.execute(tbl.select())]) for tbl in
                          metadata.sorted_tables])
        assert total_nentries == 0

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_object[Simulation] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'simulationId'
                },
                {
                    'name': 'name'
                },
                {
                    'name': 'datePosted'
                },
                {
                    'name': 'mapper'
                },
                {
                    'name': 'note'
                },
                {
                    'name': 'simulationFilePaths'
                },
                {
                    'name': 'id'
                }
            ],
            'name': 'Simulation'
        }
    }
}

snapshots['test_object[SimulationConnection] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'pageInfo'
                },
                {
                    'name': 'edges'
                }
            ],
            'name': 'SimulationConnection'
        }
    }
}

snapshots['test_object[SimulationFilePath] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'simulationFilePathId'
                },
                {
                    'name': 'simulationId'
                },
                {
                    'name': 'path'
                },
                {
                    'name': 'note'
                },
                {
                    'name': 'simulation'
                },
                {
                    'name': 'id'
                }
            ],
            'name': 'SimulationFilePath'
        }
    }
}

snapshots['test_object[SimulationFilePathConnection] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'pageInfo'
                },
                {
                    'name': 'edges'
                }
            ],
            'name': 'SimulationFilePathConnection'
        }
    }
}

snapshots['test_object[Map] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'mapId'
                },
                {
                    'name': 'name'
                },
                {
                    'name': 'datePosted'
                },
                {
                    'name': 'mapper'
                },
                {
                    'name': 'note'
                },
                {
                    'name': 'mapFilePaths'
                },
                {
                    'name': 'beams'
                },
                {
                    'name': 'id'
                }
            ],
            'name': 'Map'
        }
    }
}

snapshots['test_object[MapConnection] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'pageInfo'
                },
                {
                    'name': 'edges'
                }
            ],
            'name': 'MapConnection'
        }
    }
}

snapshots['test_object[MapFilePath] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'mapFilePathId'
                },
                {
                    'name': 'mapId'
                },
                {
                    'name': 'path'
                },
                {
                    'name': 'note'
                },
                {
                    'name': 'map'
                },
                {
                    'name': 'id'
                }
            ],
            'name': 'MapFilePath'
        }
    }
}

snapshots['test_object[MapFilePathConnection] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'pageInfo'
                },
                {
                    'name': 'edges'
                }
            ],
            'name': 'MapFilePathConnection'
        }
    }
}

snapshots['test_object[Beam] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'beamId'
                },
                {
                    'name': 'name'
                },
                {
                    'name': 'inputMapId'
                },
                {
                    'name': 'inputBeamId'
                },
                {
                    'name': 'map'
                },
                {
                    'name': 'parentBeam'
                },
                {
                    'name': 'childBeams'
                },
                {
                    'name': 'beamFilePaths'
                },
                {
                    'name': 'id'
                }
            ],
            'name': 'Beam'
        }
    }
}

snapshots['test_object[BeamConnection] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'pageInfo'
                },
                {
                    'name': 'edges'
                }
            ],
            'name': 'BeamConnection'
        }
    }
}

snapshots['test_object[BeamFilePath] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'beamFilePathId'
                },
                {
                    'name': 'beamId'
                },
                {
                    'name': 'path'
                },
                {
                    'name': 'note'
                },
                {
                    'name': 'beam'
                },
                {
                    'name': 'id'
                }
            ],
            'name': 'BeamFilePath'
        }
    }
}

snapshots['test_object[BeamFilePathConnection] 1'] = {
    'data': {
        '__type': {
            'description': None,
            'fields': [
                {
                    'name': 'pageInfo'
                },
                {
                    'name': 'edges'
                }
            ],
            'name': 'BeamFilePathConnection'
        }
    }
}



"""the instance of SQLAlchemy from Flask-SQLAlchemy

"""
from flask_sqlalchemy import SQLAlchemy

##__________________________________________________________________||
sa = SQLAlchemy()
"""the instance of SQLAlchemy from Flask-SQLAlchemy

The instance is named "sa" here. However, it is more commonly named
"db" in example code found online.

- API: https://flask-sqlalchemy.palletsprojects.com/en/2.x/api/#flask_sqlalchemy.SQLAlchemy
- The difference from the plain SQLAlchemy:
  https://flask-sqlalchemy.palletsprojects.com/en/2.x/api/#flask_sqlalchemy.SQLAlchemy

"""
##__________________________________________________________________||



from pathlib import Path
from flask_migrate import Migrate

from .sa import sa
from .conn import close_db_connection
from .cmds import init_db_command
from .cmds import dump_db_command
from .cmds import import_csv_command
from .cmds import export_csv_command
from .cmds import backup_db_command

migrate = Migrate()

_MIGRATIONS_DIR = Path(__file__).resolve().parent.parent.joinpath('migrations')

##__________________________________________________________________||
def init_app(app):
    """initializes the Flask application object

    This function is called by `create_app()` of Flask

    Parameters
    ----------
    app : Flask
        The Flask application object, an instance of `Flask`
    """
    sa.init_app(app)
    migrate.init_app(app, sa, directory=_MIGRATIONS_DIR)
    app.cli.add_command(init_db_command)
    app.cli.add_command(dump_db_command)
    app.cli.add_command(import_csv_command)
    app.cli.add_command(export_csv_command)
    app.cli.add_command(backup_db_command)
    app.teardown_appcontext(close_db_connection)

##__________________________________________________________________||



from pathlib import Path
import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables, import_tables_from_csv_files
from acondbs.db.ops import export_db_to_dict_of_dict_list
from acondbs.db.ops import export_db_to_csv_files

from ...constants import SAMPLE_DIR

##__________________________________________________________________||
@pytest.fixture
def app():
    database_uri ="sqlite:///:memory:"
    csvdir = Path(SAMPLE_DIR, 'csv')
    app = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with app.app_context():
        define_tables()
        import_tables_from_csv_files(csvdir)
    yield app

##__________________________________________________________________||
def test_export_db_to_csv_files(app, tmpdir_factory, snapshot):
    outdir = str(tmpdir_factory.mktemp('csv'))

    with app.app_context():
        export_db_to_csv_files(outdir)

    with app.app_context():
        define_tables()
        import_tables_from_csv_files(outdir)
        snapshot.assert_match(export_db_to_dict_of_dict_list())

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[updateMapFilePath] 1'] = {
    'data': {
        'updateMapFilePath': {
            'mapFilePath': {
                'path': 'nersc:/go/to/my/new_map_v2'
            }
        }
    }
}

snapshots['test_schema_success[updateMapFilePath] 2'] = {
    'data': {
        'map': {
            'beams': {
                'edges': [
                    {
                        'node': {
                            'name': '20200123'
                        }
                    }
                ]
            },
            'datePosted': '2020-01-20',
            'mapFilePaths': {
                'edges': [
                    {
                        'node': {
                            'map': {
                                'mapId': '1012'
                            },
                            'note': '- Note 1 updated',
                            'path': 'nersc:/go/to/my/new_map_v2'
                        }
                    },
                    {
                        'node': {
                            'map': {
                                'mapId': '1012'
                            },
                            'note': 'lat only',
                            'path': 'nersc:/go/to/my/maps_v2'
                        }
                    },
                    {
                        'node': {
                            'map': {
                                'mapId': '1012'
                            },
                            'note': 'lat only',
                            'path': 'abcde:/path/to/the/maps_v2'
                        }
                    }
                ]
            },
            'mapper': 'pwg-pmn',
            'name': 'lat20200120',
            'note': '''- This is a dummy test with a lat map
- A beam depends on this map'''
        }
    }
}



import git
from pathlib import Path

import pytest

from acondbs.misc import gitb

##__________________________________________________________________||
def test_pull(remote_url, tmpdir_factory):
    """test pull()

    """

    # create two clones
    folder1 = Path(tmpdir_factory.mktemp('git'))
    repo1 = git.Repo.clone_from(remote_url, folder1)

    folder2 = Path(tmpdir_factory.mktemp('git'))
    repo2 = git.Repo.clone_from(remote_url, folder2)

    # push a commit from repo1
    file1 = folder1.joinpath('f.txt')
    with file1.open("a") as f:
        f.write('1')
    repo1.git.add(A=True)
    repo1.index.commit('update')
    remote1 = repo1.remotes[repo1.active_branch.tracking_branch().remote_name]
    remote1.push()
    head_sha_repo1 = repo1.head.commit.hexsha

    # save sha
    head_sha_old = repo2.head.commit.hexsha

    # call pull()
    gitb.pull(folder2)

    # assert
    head_sha_new = repo2.head.commit.hexsha
    assert not head_sha_old == head_sha_new
    assert head_sha_repo1 == head_sha_new

##__________________________________________________________________||
def test_nonexistent_path(tmpdir_factory):
    """assert exception is raised for nonexistent path
    """
    folder = Path(tmpdir_factory.mktemp('git'))
    path = folder.joinpath('nonexistent')

    with pytest.raises(ValueError):
        gitb.pull(path)

def test_path_not_repo(folder):
    """assert exception is raised if not a repo
    """

    with pytest.raises(ValueError):
        gitb.pull(folder)

def test_repo_no_remote(repo):
    """assert exception is raised if a repo has no tracking branch
    """

    with pytest.raises(ValueError):
        gitb.pull(repo.working_tree_dir)

##__________________________________________________________________||



from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import SimulationFilePath as SimulationFilePathModel

##__________________________________________________________________||
class SimulationFilePath(SQLAlchemyObjectType):
    class Meta:
        model = SimulationFilePathModel
        interfaces = (relay.Node, )

##__________________________________________________________________||



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import MapFilePath as MapFilePathModel

from ..db.sa import sa
from ..db.backup import request_backup_db

##__________________________________________________________________||
class MapFilePath(SQLAlchemyObjectType):
    class Meta:
        model = MapFilePathModel
        interfaces = (relay.Node, )

##__________________________________________________________________||
class MapFilePathAttribute:
    path = graphene.String()
    note = graphene.String()
    map_id = graphene.Int()

class CreateMapFilePathInput(graphene.InputObjectType, MapFilePathAttribute):
    pass

class UpdateMapFilePathInput(graphene.InputObjectType, MapFilePathAttribute):
    pass

class CreateMapFilePath(graphene.Mutation):
    class Arguments:
        input = CreateMapFilePathInput(required=True)

    ok = graphene.Boolean()
    mapFilePath = graphene.Field(lambda: MapFilePath)

    def mutate(root, info, input):
        mapFilePath = MapFilePathModel(**input)
        sa.session.add(mapFilePath)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateMapFilePath(mapFilePath=mapFilePath, ok=ok)

class UpdateMapFilePath(graphene.Mutation):
    class Arguments:
        map_file_path_id = graphene.Int()
        input = UpdateMapFilePathInput(required=True)

    ok = graphene.Boolean()
    mapFilePath = graphene.Field(lambda: MapFilePath)

    def mutate(root, info, map_file_path_id, input):
        mapFilePath = MapFilePathModel.query.filter_by(map_file_path_id=map_file_path_id).first()
        for k, v in input.items():
            setattr(mapFilePath, k, v)
        sa.session.commit()
        ok = True
        request_backup_db()
        return UpdateMapFilePath(mapFilePath=mapFilePath, ok=ok)

class DeleteMapFilePath(graphene.Mutation):
    class Arguments:
        map_file_path_id = graphene.Int()

    ok = graphene.Boolean()

    def mutate(root, info, map_file_path_id):
        mapFilePath = MapFilePathModel.query.filter_by(map_file_path_id=map_file_path_id).first()
        sa.session.delete(mapFilePath)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteMapFilePath(ok=ok)

##__________________________________________________________________||



from pathlib import Path


import pytest
import unittest.mock as mock

from acondbs.misc.lock import lock
from acondbs.misc.lock import TimeOutAcquiringLock

##__________________________________________________________________||
def test_acquire_release(tmpdir_factory):
    folder = Path(tmpdir_factory.mktemp('lock'))
    lock_file = folder.joinpath('.lock')

    l = lock(lock_file)
    assert not l.locked
    l.acquire()
    assert l.locked
    assert lock_file.exists()
    l.release()
    assert not l.locked
    assert not lock_file.exists()

##__________________________________________________________________||
def test_release_when_not_lockded(tmpdir_factory):
    folder = Path(tmpdir_factory.mktemp('lock'))
    lock_file = folder.joinpath('.lock')

    l = lock(lock_file)
    assert not l.locked
    l.release()
    assert not l.locked
    assert not lock_file.exists()

##__________________________________________________________________||
def test_release_not_delete_when_not_lockded(tmpdir_factory):
    folder = Path(tmpdir_factory.mktemp('lock'))
    lock_file = folder.joinpath('.lock')

    lock_file.touch()

    l = lock(lock_file)
    assert not l.locked
    l.release()
    assert not l.locked
    assert lock_file.exists()

##__________________________________________________________________||
def test_acquire_timeout(tmpdir_factory):
    timeout = 0.1 # sec

    folder = Path(tmpdir_factory.mktemp('lock'))
    lock_file = folder.joinpath('.lock')

    lock_file.touch()

    l = lock(lock_file, timeout=timeout)

    with pytest.raises(TimeOutAcquiringLock):
        l.acquire()

##__________________________________________________________________||
def test_with_success(tmpdir_factory):
    folder = Path(tmpdir_factory.mktemp('lock'))
    lock_file = folder.joinpath('.lock')

    with lock(lock_file) as l:
        assert l.locked
        assert lock_file.exists()
    assert not l.locked
    assert not lock_file.exists()

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[allMapsFirstTwo] 1'] = {
    'data': {
        'allMaps': {
            'edges': [
                {
                    'node': {
                        'name': 'lat20190213'
                    }
                },
                {
                    'node': {
                        'name': 'lat20200120'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[allMapsFirstTwoSort] 1'] = {
    'data': {
        'allMaps': {
            'edges': [
                {
                    'node': {
                        'name': 'lat20200201'
                    }
                },
                {
                    'node': {
                        'name': 'lat20200120'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[mapByMapID] 1'] = {
    'data': {
        'map': {
            'name': 'lat20190213'
        }
    }
}

snapshots['test_schema[mapByMapID-nonexistent] 1'] = {
    'data': {
        'map': None
    }
}

snapshots['test_schema[mapByName] 1'] = {
    'data': {
        'map': {
            'mapId': '1001'
        }
    }
}



import pytest
from graphene.test import Client

from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    'Simulation', 'SimulationConnection',
    'SimulationFilePath', 'SimulationFilePathConnection',
    'Map', 'MapConnection',
    'MapFilePath', 'MapFilePathConnection',
    'Beam', 'BeamConnection',
    'BeamFilePath', 'BeamFilePathConnection',
]

@pytest.mark.parametrize('type_name', params)
def test_object(app, snapshot, type_name):
    '''test objects defined
    '''
    query = '''
      query q($type: String!) {
        __type(name: $type) {
          name
          description
          fields {
            name
          }
        }
      }
    '''
    variables = {'type': type_name}
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(query, variables=variables))

##__________________________________________________________________||



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyConnectionField

from .simulation import Simulation, SimulationModel
from .map_ import Map, MapModel, MapConnection
from .beam import Beam, BeamModel
from .beam_file_path import BeamFilePath, BeamFilePathModel
from .map_file_path import MapFilePath, MapFilePathModel
from .simulation_file_path import SimulationFilePath, SimulationFilePathModel

##__________________________________________________________________||
class Query(graphene.ObjectType):

    version = graphene.String()

    def resolve_version(self, info):
        from .. import __version__
        return __version__

    node = relay.Node.Field()
    all_simulations = SQLAlchemyConnectionField(Simulation._meta.connection)
    all_maps = SQLAlchemyConnectionField(MapConnection)
    all_beams = SQLAlchemyConnectionField(Beam._meta.connection)
    all_simulation_file_paths = SQLAlchemyConnectionField(SimulationFilePath._meta.connection)
    all_map_file_paths = SQLAlchemyConnectionField(MapFilePath._meta.connection)
    all_beam_file_paths = SQLAlchemyConnectionField(BeamFilePath._meta.connection)

    simulation = graphene.Field(Simulation, simulation_id=graphene.Int(), name=graphene.String())

    def resolve_simulation(self, info, **kwargs):
        import time, random
        print(kwargs)
        # time.sleep(random.randint(1, 5))
        fields = ('simulation_id', 'name')
        query = Simulation.get_query(info)
        for f in fields:
            v = kwargs.get(f)
            if v is not None:
                query = query.filter(getattr(SimulationModel, f)==v)
                return query.first()
        return None

    map = graphene.Field(Map, map_id=graphene.Int(), name=graphene.String())

    def resolve_map(self, info, **kwargs):
        import time, random
        print(kwargs)
        # time.sleep(random.randint(1, 5))
        fields = ('map_id', 'name')
        query = Map.get_query(info)
        for f in fields:
            v = kwargs.get(f)
            if v is not None:
                query = query.filter(getattr(MapModel, f)==v)
                return query.first()
        return None

    beam = graphene.Field(Beam, beam_id=graphene.Int(), name=graphene.String())

    def resolve_beam(self, info, **kwargs):
        fields = ('beam_id', 'name')
        query = Beam.get_query(info)
        for f in fields:
            v = kwargs.get(f)
            if v is not None:
                query = query.filter(getattr(BeamModel, f)==v)
                return query.first()
        return None
##__________________________________________________________________||



"""Throttle function executions
"""
import threading
import functools
import queue
import time
from enum import Enum

##__________________________________________________________________||
class cap_exec_rate:
    """cap the execution rate of a function

    This class prevents a function from being executed too often. This
    class executes a function in another thread. Once it executes a
    function, it starts a timer. If the function is called again
    through this class while the timer is still running, this class
    does not execute the function immediately. Instead, this class
    executes the function after the timer goes off.

    The function might be called multiple times through this class
    while the timer is running. After the timer goes off, this class
    only executes the function once.

    Parameters
    ----------
    func : callable
        A function or any callable object to be executed not too often
    pause_time : float
        The length of time [second] to be waited after an execution of
        the `func` before the next execution. If the `func` is called
        through this class once or more times within this time length
        after an execution, it will be executed once after this time
        length passes.
    daemon : bool
        whether the thread which executes the function and sets the
        timer is daemonic. If `False`, the Python program doesn't exit
        until the thread ends; the method `end()` needs to be called;
        the function will be executed one last time if it has been
        called but not executed. If `True`, the thread can be abruptly
        stopped at the end of Python program; the function that has
        been called but not executed will not be executed.

    """
    def __init__(self, func, pause_time=1.0, daemon=False):
        self.func = func
        self.pause_time = pause_time
        self.daemon = daemon

        self.queue = queue.Queue()
        config = Config(func, self.queue, pause_time)
        self.machine = threading.Thread(
            target=state_machine, args=(config, ), daemon=daemon)
        self.machine.start()

    def __call__(self):
        """call the function

        The number of executions of the function is capped
        """
        self.queue.put(Message.FUNC_CALLED)

    def end(self):
        """end using this class

        The function will be executed immediately if it is waiting for
        the timer. This method returns after the function is executed
        and the thread joins.

        """
        self.queue.put(Message.EXIT)
        self.machine.join()

def state_machine(config):
    """the entry function to be executed in a thread
    """
    state = Active(config)
    while not state.end:
        state = state()

class Message(Enum):
    FUNC_CALLED = 1
    TIMER_GOES_OFF = 2
    EXIT = 3

class Config:
    """Configuration of the state
    """
    def __init__(self, func, queue, pause_time):
        self.func = func
        self.queue = queue
        self.pause_time = pause_time

##__________________________________________________________________||
class State:
    """The base class of the states
    """
    def __init__(self, config):
        self.config = config
    def __call__(self):
        message = self.config.queue.get()
        if message == Message.FUNC_CALLED:
            return self.func_called()
        elif message == Message.TIMER_GOES_OFF:
            return self.timer_goes_off()
        elif message == Message.EXIT:
            return self.exit_()
        else:
            raise ValueError('unknown message: {!r}'.format(message))

class Active(State):
    """Active state

    The timer is not running. The function can be executed immediately
    if it is called.
    """
    end = False
    def __init__(self, config):
        super().__init__(config)
    def func_called(self):
        self.config.func()
        return Pause(self.config)
    def exit_(self):
        return Exit(self.config)

def timer_end(queue):
    queue.put(Message.TIMER_GOES_OFF)

class Pause(State):
    """Pause state. The function has not been called

    The timer is running. The function has not been called. It won't
    be executed immediately when it is called.
    """
    end = False
    def __init__(self, config):
        super().__init__(config)
        self.timer = threading.Timer(
            self.config.pause_time,
            functools.partial(timer_end, self.config.queue))
        self.timer.start()
    def func_called(self):
        return PauseCalled(self.config, self.timer)
    def timer_goes_off(self):
        return Active(self.config)
    def exit_(self):
        self.timer.cancel()
        return Exit(self.config)

class PauseCalled(State):
    """Pause state. The function has been called

    The timer is running. The function has been called. The function
    won't be executed until the timer goes off.
    """
    end = False
    def __init__(self, config, timer):
        super().__init__(config)
        self.timer = timer
    def func_called(self):
        return self
    def timer_goes_off(self):
        self.config.func()
        return Pause(self.config)
    def exit_(self):
        self.timer.cancel()
        self.config.func()
        return Exit(self.config)

class Exit(State):
    """Exit state
    """
    end = True
    def __init__(self, config):
        super().__init__(config)

##__________________________________________________________________||



from setuptools import setup, find_packages
import versioneer

from pathlib import Path

here = Path(__file__).resolve().parent
long_description = here.joinpath('README.md').read_text()

setup(
    name='acondbs',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='A GraphQL server for product DB',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Simons Observatory',
    author_email='so_software@simonsobservatory.org',
    url='https://github.com/simonsobs/acondbs',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'Flask>=1.1',
        'Flask-Cors>=3.0',
        'Flask-GraphQL>=2.0',
        'Flask-Migrate>=2.5',
        'Flask-SQLAlchemy>=2.4',
        'graphene-sqlalchemy>=2.2',
        'gitpython>=3.1'
    ],
    extras_require={
        'tests': [
            'pytest>-5.4',
            'pytest-cov>=2.8',
            'snapshottest>0.5'
        ]
    }
)



import pytest
from graphene.test import Client

from acondbs.schema.schema import schema

##__________________________________________________________________||
params = [
    pytest.param(
        '''
        { allMaps(first: 2) {
             edges { node { name } }
           } }
         ''',
        id='allMapsFirstTwo'
    ),
    pytest.param(
        '''
        { allMaps(first: 2, sort: DATE_POSTED_DESC) {
             edges { node { name } }
           } }
         ''',
        id='allMapsFirstTwoSort'
    ),
    pytest.param(
        '''
        { map(mapId: 1001) { name } }
         ''',
        id='mapByMapID'
    ),
    pytest.param(
        '''
        { map(mapId: 2001) { name } }
         ''',
        id='mapByMapID-nonexistent'
    ),
    pytest.param(
        '''
        { map(name: "lat20190213") { mapId } }
         ''',
        id='mapByName'
    ),
]

@pytest.mark.parametrize('query', params)
def test_schema(app, snapshot, query):
    client = Client(schema)
    with app.app_context():
        snapshot.assert_match(client.execute(query))

##__________________________________________________________________||



from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Beam as BeamModel

##__________________________________________________________________||
class Beam(SQLAlchemyObjectType):
    class Meta:
        model = BeamModel
        interfaces = (relay.Node, )

# class BeamConnection(relay.Connection):
#      class Meta:
#          node = Beam

## The class "BeamConnection" is commented out because it causes the error
## "AssertionError: Found different types with the same name in the schema:
## BeamConnection, BeamConnection". In the class "Query" in schema.py
## "Beam._meta.connection" is used instead. This solution was mentioned in the
## issue
## https://github.com/graphql-python/graphene-sqlalchemy/issues/153#issuecomment-478744077

##__________________________________________________________________||



from pathlib import Path
import git

import pytest

##__________________________________________________________________||
@pytest.fixture()
def folder(tmpdir_factory):
    """path to an folder (not a git repo) with two text files

    f.txt
       abc

    g.txt
       123

    """
    folder = Path(tmpdir_factory.mktemp('git'))
    file1 = folder.joinpath('f.txt')
    file2 = folder.joinpath('g.txt')
    file1.write_text('abc\n')
    file2.write_text('123\n')
    yield folder

##__________________________________________________________________||
@pytest.fixture()
def repo(folder):
    """a git repo

    The folder is initialized as a git repo.
    The two files in the folder are committed.


    """
    repo = git.Repo.init(folder)
    repo.git.add(A=True)
    repo.index.commit('initial commit')
    yield repo

##__________________________________________________________________||
@pytest.fixture()
def bare_repo(repo, tmpdir_factory):
    """a bare repo, a clone of the repo
    """
    folder = Path(tmpdir_factory.mktemp('git'))
    y = git.Repo.clone_from(repo.working_tree_dir, folder, bare=True)
    yield y

@pytest.fixture()
def github_repo_url(bare_repo, tmpdir_factory):
    """a factory fixture
    """
    url = 'git@github.com:TaiSakuma/fuzzy-garbanzo.git'
    def _f():
        """create a clone of the bare repo and push force
           the content to the github repo

        """
        folder = Path(tmpdir_factory.mktemp('git'))
        clone = git.Repo.clone_from(bare_repo.git_dir, folder)
        branch_name = clone.active_branch.name
        remote = clone.create_remote('github', url=url)
        remote.push(refspec='{}:{}'.format(branch_name, branch_name), force=True)
        return url
    yield _f

@pytest.fixture(params=[
    'local_folder',
    # 'github_repo'
])
def remote_url(request, bare_repo, github_repo_url):
    p = request.param
    if p == 'local_folder':
        y = bare_repo.git_dir
    elif p == 'github_repo':
        y = github_repo_url()
    else:
        raise ValueError('unknown param: {}'.format(p))
    yield y

##__________________________________________________________________||
